@@ -1,10 +1,4 @@
-# -*- coding: utf-8 -*-
-# Copyright (C) 2016, the Pyzo development team
-#
-# Pyzo is distributed under the terms of the 2-Clause BSD License.
-# The full license can be found in 'license.txt'.
-
-""" Module codeparser
+"""Module codeparser
 
 Analyses the source code to get the structure of a module/script.
 This can be used for fictive introspection, and to display the
@@ -14,7 +8,10 @@ structure of a source file in for example a tree widget.
 
 # TODO: replace this module, get data from the syntax highlighter in the code editor
 
-import time, threading, re
+from collections import deque
+import time
+import threading
+import re
 import pyzo
 
 
@@ -53,8 +50,8 @@ class Result:
         self.editorId = editorId
 
     def isMatch(self, editorId):
-        """isMatch(editorId):
-        Returns whether the result matches with the given editorId.
+        """Returns whether the result matches with the given editorId.
+
         The editorId can also be an editor instance."""
         if isinstance(editorId, int):
             return self.editorId == editorId
@@ -65,24 +62,24 @@ class Result:
 class Parser(threading.Thread):
     """Parser
     Parsing sourcecode in a separate thread, this class obtains
-    introspection informarion. This class is also the interface
+    introspection information. This class is also the interface
     to the parsed information; it has methods that can be used
     to extract information from the result.
     """
 
     def __init__(self):
-        threading.Thread.__init__(self)
+        super().__init__()
 
         # Reference current job
         self._job = None
 
         # Reference to last result
-        self._result = None
+        self._results = deque([], 10)
 
         # Lock to enable save threading
         self._lock = threading.RLock()
 
-        # Set deamon
+        # Set daemon
         self.daemon = True
         self._exit = False
 
@@ -91,8 +88,8 @@ class Parser(threading.Thread):
         self.join(timeout)
 
     def parseThis(self, editor):
-        """parseThis(editor)
-        Give the parser new text to parse.
+        """Give the parser new text to parse.
+
         If the parser is busy parsing text, it will stop doing that
         and start anew with the most recent version of the text.
         """
@@ -101,18 +98,17 @@ class Parser(threading.Thread):
         text = editor.toPlainText()
 
         # Make job
-        self._lock.acquire()
-        self._job = Job(text, id(editor))
-        self._lock.release()
+        with self._lock:
+            self._job = Job(text, id(editor))
 
     def getFictiveNameSpace(self, editor):
-        """getFictiveNameSpace(editor)
-        Produce the fictive namespace, based on the current position.
+        """Produce the fictive namespace, based on the current position.
+
         A list of names is returned.
         """
 
         # Obtain result
-        result = self._getResult()
+        result = self._getResult(editor)
         if result is None or not result.isMatch(editor):
             return []
 
@@ -153,17 +149,16 @@ class Parser(threading.Thread):
         return namespace
 
     def getFictiveClass(self, name, editor, handleSelf=False):
-        """getFictiveClass(name, editor, handleSelf=False)
-        Return the fictive class object of the given name, or None
+        """return the fictive class object of the given name, or None
         if it does not exist. If handleSelf is True, automatically
         handles "self." names.
         """
         return self._getFictiveItem(name, "class", editor, handleSelf)
 
     def getFictiveSignature(self, name, editor, handleSelf=False):
-        """getFictiveSignature(name, editor, handleSelf=False)
-        Get the signature of the fictive function or method of the
-        given name. Returns None if the given name is not a known
+        """get the signature of the fictive function or method of the given name
+
+        Returns None if the given name is not a known
         function or method. If handleSelf is True, automatically
         handles "self." names.
         """
@@ -189,15 +184,15 @@ class Parser(threading.Thread):
             return None
 
     def getFictiveImports(self, editor):
-        """getFictiveImports(editor)
-        Get the fictive imports of this source file.
+        """get the fictive imports of this source file.
+
         tuple:
         - list of names that are imported,
         - a dict with the line to import each name
         """
 
         # Obtain result
-        result = self._getResult()
+        result = self._getResult(editor)
         if result is None or not result.isMatch(editor):
             return [], []
 
@@ -209,24 +204,26 @@ class Parser(threading.Thread):
             importlines[item.name] = item.text
         return imports, importlines
 
-    def _getResult(self):
-        """getResult()
-        Savely Obtain result.
-        """
-        self._lock.acquire()
-        result = self._result
-        self._lock.release()
+    def _getResult(self, editor):
+        """safely Obtain result"""
+        editorId = id(editor)
+        result = None
+        with self._lock:
+            for res in self._results:
+                if res.editorId == editorId:
+                    result = res
+                    break
         return result
 
     def _getFictiveItem(self, name, type, editor, handleSelf=False):
-        """_getFictiveItem(name, type, editor, handleSelf=False)
-        Obtain the fictive item of the given name and type.
+        """Obtain the fictive item of the given name and type.
+
         If handleSelf is True, will handle "self." correctly.
         Intended for internal use.
         """
 
         # Obtain result
-        result = self._getResult()
+        result = self._getResult(editor)
         if result is None or not result.isMatch(editor):
             return None
 
@@ -267,14 +264,13 @@ class Parser(threading.Thread):
         return theItem
 
     def _getFictiveCurrentClass(self, editor, selfname):
-        """_getFictiveCurrentClass(editor, selfname)
-        Get the fictive object for the class referenced
+        """get the fictive object for the class referenced
         using selfname (usually 'self').
-        Intendef for internal use.
+        Intended for internal use.
         """
 
         # Obtain result
-        result = self._getResult()
+        result = self._getResult(editor)
         if result is None:
             return None
 
@@ -310,9 +306,7 @@ class Parser(threading.Thread):
         return theclass
 
     def run(self):
-        """run()
-        This is the main loop.
-        """
+        """This is the main loop."""
 
         time.sleep(0.5)
         try:
@@ -323,20 +317,17 @@ class Parser(threading.Thread):
                     return
 
                 if self._job:
-
-                    # Savely obtain job
-                    self._lock.acquire()
-                    job = self._job
-                    self._job = None
-                    self._lock.release()
+                    # Safely obtain job
+                    with self._lock:
+                        job = self._job
+                        self._job = None
 
                     # Analyse job
                     result = self._analyze(job)
 
-                    # Savely store result
-                    self._lock.acquire()
-                    self._result = result
-                    self._lock.release()
+                    # Safely store result
+                    with self._lock:
+                        self._results.appendleft(result)
 
                     # Notify
                     if pyzo.editors is not None:
@@ -376,10 +367,10 @@ class Parser(threading.Thread):
 
         # To know when to make something new when for instance a class is defined
         # in an if statement, we keep track of the last valid node/object:
-        # Put inside a list, so we can set it from inside a subfuncion
+        # Put inside a list, so we can set it from inside a subfunction
         lastObject = [root]
 
-        # Define funcion to put an item in the structure in the right parent
+        # Define function to put an item in the structure in the right parent
         def appendToStructure(object):
             # find position in structure to insert
             node = lastObject[0]
@@ -393,11 +384,7 @@ class Parser(threading.Thread):
 
         # Find objects!
         # type can be: cell, class, def, import, var
-        for i in range(len(lines)):
-
-            # Obtain line
-            line = lines[i]
-
+        for i, line in enumerate(lines):
             # Should we stop?
             if self._job or self._exit:
                 break
@@ -408,17 +395,12 @@ class Parser(threading.Thread):
             line = tmp.rstrip()
 
             # Detect cells
-            if (
-                line.startswith("##")
-                or line.startswith("#%%")
-                or line.startswith("# %%")
-            ):
-                if line.startswith("##"):
-                    name = line[2:].lstrip()
-                elif line.startswith("#%%"):
-                    name = line[3:].lstrip()
-                else:
-                    name = line[4:].lstrip()
+            cellPrefixes = ("##", "#%%", "# %%")
+            if line.startswith(cellPrefixes):
+                for s in cellPrefixes:
+                    if line.startswith(s):
+                        name = line[len(s) :].lstrip()
+                        break
                 item = FictiveObject("cell", i, indent, name)
                 leafs.append(item)
                 # Next! (we have to put this before the elif stuff below
@@ -427,10 +409,11 @@ class Parser(threading.Thread):
 
             # Split in line and comment
             line, tmp, cmnt = line.partition("#")
-            line, cmnt = line.rstrip(), cmnt.lower().strip()
+            line, cmnt = line.rstrip(), cmnt.strip()
 
             # Detect todos
-            if cmnt and (cmnt.startswith("todo:") or cmnt.startswith("2do:")):
+            firstWord = cmnt.lstrip().split(" ", 1)[0].rstrip(":")
+            if firstWord.lower() in ["todo", "2do", "fixme"]:
                 item = FictiveObject("todo", i, indent, cmnt)
                 item.linenr2 = i + 1  # a todo is active at one line only
                 leafs.append(item)
@@ -558,8 +541,7 @@ class Parser(threading.Thread):
             # find object after linenr
             object1, object2 = None, None  # if no items at all
             i = -1
-            for i in range(len(series)):
-                object = series[i]
+            for i, object in enumerate(series):
                 if object.type not in ["class", "def"]:
                     continue
                 if object.linenr > linenr:
@@ -639,20 +621,11 @@ class FictiveObject:
         self.sig = ""  # for functions and methods
 
 
-namechars = "abcdefghijklmnopqrstuvwxyz_0123456789"
-
-
 def IsValidName(name):
     """Given a string, checks whether it is a
     valid name (dots are not valid!)
     """
-    if not name:
-        return False
-    name = name.lower()
-    if name[0] not in namechars[0:-10]:
-        return False
-    tmp = map(lambda x: x not in namechars, name[2:])
-    return sum(tmp) == 0
+    return name.isidentifier()
 
 
 def ParseImport(names):
@@ -666,9 +639,10 @@ def ParseImport(names):
 
 
 def findString(text, s, i):
-    """findString(text, s)
-    Find s in text, but only if s is not in a string or commented
-    Helper function for washMultilineStrings"""
+    """find s in text, but only if s is not in a string or commented
+
+    Helper function for washMultilineStrings
+    """
 
     while True:
         i = _findString(text, s, i)
@@ -697,7 +671,7 @@ def _findString(text, s, i):
     line = text[i1:i2]
 
     # Count quotes, we're done if we found none
-    if not line.count('"') and not line.count("'") and not line.count("#"):
+    if '"' not in line and "'" not in line and "#" not in line:
         return i2
 
     # So we found quotes, now really count them ...
@@ -716,7 +690,7 @@ def _findString(text, s, i):
                 if inString == c:
                     inString = ""  # exit string
                 else:
-                    pass  # the other quote can savely be used inside this string
+                    pass  # the other quote can safely be used inside this string
         prev = c
 
     # If we are in a string, this match is false ...
@@ -727,8 +701,7 @@ def _findString(text, s, i):
 
 
 def washMultilineStrings(text):
-    """washMultilineStrings(text)
-    Replace all text within multiline strings with dummy chars
+    """Replace all text within multiline strings with dummy chars
     so that it is not parsed.
     """
     i = 0
@@ -744,9 +717,9 @@ def washMultilineStrings(text):
         else:
             # Make no result be very large
             if i1 == -1:
-                i1 = 2 ** 60
+                i1 = 2**60
             if i2 == -1:
-                i2 = 2 ** 60
+                i2 = 2**60
             # Find end of the multiline comment
             if i1 < i2:
                 i3 = i1 + 3
@@ -756,7 +729,7 @@ def washMultilineStrings(text):
                 i4 = text.find(s2, i3)
             # No end found -> take all text, unclosed string!
             if i4 == -1:
-                i4 = 2 ** 32
+                i4 = 2**32
             # Leave only the first two quotes of the start of the comment
             i3 -= 1
             i4 += 3
