@@ -1,9 +1,4 @@
-# Copyright (C) 2016, the Pyzo development team
-#
-# Pyzo is distributed under the terms of the 2-Clause BSD License.
-# The full license can be found in 'license.txt'.
-
-""" Module main
+"""Module main
 
 This module contains the main frame. Implements the main window.
 Also adds some variables to the pyzo namespace, such as the callLater
@@ -11,7 +6,9 @@ function which is also defined here.
 
 """
 
-import os, sys, time
+import os
+import sys
+import time
 import base64
 from queue import Queue, Empty
 
@@ -539,6 +536,10 @@ def loadIcons():
     artist.addLayer("arrow_refresh")
     pyzo.icons["reload_file_from_disk"] = artist.finish()
 
+    artist = IconArtist("page_white_copy")
+    artist.addLayer("overlay_disk")
+    pyzo.icons["save_copy_as"] = artist.finish()
+
 
 def loadFonts():
     """Load all fonts that come with Pyzo."""
