@@ -1,11 +1,4 @@
-# -*- coding: utf-8 -*-
-# Copyright (C) 2016, the Pyzo development team
-#
-# Pyzo is distributed under the terms of the 2-Clause BSD License.
-# The full license can be found in 'license.txt'.
-
-
-""" Module pyzokernel.interpreter
+"""Module pyzokernel.interpreter
 
 Implements the Pyzo interpreter.
 
@@ -24,6 +17,7 @@ We integrate IPython via the IPython.core.interactiveshell.InteractiveShell.
 """
 
 import os
+import re
 import sys
 import time
 import logging
@@ -59,6 +53,9 @@ else:
     bstr = bytes
 
 
+file_linenr_pattern = re.compile(r"(?:\.py|<tmp \d+>)(\+\d+)['\",]+ line (\d+)")
+
+
 class PS1:
     """Dynamic prompt for PS1. Show IPython prompt if available, and
     show current stack frame when debugging.
@@ -224,12 +221,21 @@ class PyzoInterpreter:
             del tb
             if self._exitException is None:
                 self._exitException = value
+        finally:
+            self._cleanup()
 
         # Exit
         if self._exitException is None:
             self._exitException = SystemExit()
         raise self._exitException
 
+    def _cleanup(self):
+        # restore prompts
+        if self._original_ps1 is not None:
+            # sys.ps1 and sys.ps2 are only defined in interactive mode
+            sys.ps1 = self._original_ps1
+            sys.ps2 = self._original_ps2
+
     def _prepare(self):
         """Prepare for running the main loop.
         Here we do some initialization like obtaining the startup info,
@@ -317,6 +323,8 @@ class PyzoInterpreter:
                 startup_info["ipython"] = "no"
 
         # Set prompts
+        self._original_ps1 = getattr(sys, "ps1", None)
+        self._original_ps2 = getattr(sys, "ps2", None)
         sys.ps1 = PS1(self)
         sys.ps2 = PS2(self)
 
@@ -361,8 +369,6 @@ class PyzoInterpreter:
                 """extracts the leading number parts of a version string to a tuple
                 e.g.: "123.45ew6.7x.dev8" --> (123, 45, 7)
                 """
-                import re
-
                 return tuple(
                     int(s) for s in re.findall(r"\.(\d+)", "." + version_string)
                 )
@@ -373,30 +379,7 @@ class PyzoInterpreter:
 
                 _nope.nope()
 
-        # Setup post-mortem debugging via appropriately logged exceptions
-        class PMHandler(logging.Handler):
-            def emit(self, record):
-                if record.exc_info:
-                    sys.last_type, sys.last_value, sys.last_traceback = record.exc_info
-                return record
-
-        # Setup logging
-        root_logger = logging.getLogger()
-        if not root_logger.handlers:
-            root_logger.addHandler(logging.StreamHandler())
-        root_logger.addHandler(PMHandler())
-
-        # Warn when logging.basicConfig is used (see issue #645)
-        def basicConfigDoesNothing(*args, **kwargs):
-            logging.warning(
-                "Pyzo already added handlers to the root handler, "
-                + "so logging.basicConfig() does nothing."
-            )
-
-        try:
-            logging.basicConfig = basicConfigDoesNothing
-        except Exception:
-            pass
+        self._add_postmortem_for_logged_exceptions()
 
         # Setup pausing of running code using SIGFPE
         signal.signal(signal.SIGFPE, self._handle_sigfpe)
@@ -404,6 +387,33 @@ class PyzoInterpreter:
         # Update startup info
         self.context._stat_startup.send(startup_info)
 
+    def _add_postmortem_for_logged_exceptions(self):
+        # Setup post-mortem debugging via appropriately logged exceptions.
+        #
+        # The Logger.callHandlers look like a good place to Monkey-patch, as
+        # it's certain that the record is going to be handled at this point,
+        # and it does not interfere with the rest of the logging mechanics.
+        #
+        # Adding a handler to the root logger (as we did earlier) seems like
+        # an elegant approach, but it gets complicated as to whether Pyzo should
+        # take responsibility for printing the message. The fact that a handler
+        # is present, will stop the logging module from printing the message,
+        # because it considers it handled. We can detect whether our handler
+        # is the only one and print if it is, but the log action may be
+        # initiated with a "higher-level" logger, so this may cause the message
+        # to be printed twice.
+
+        def callHandlers_patched(self, record):
+            if record.exc_info:
+                sys.last_type, sys.last_value, sys.last_traceback = record.exc_info
+            return callHandlers_ori(self, record)
+
+        try:
+            callHandlers_ori = logging.Logger.callHandlers
+            logging.Logger.callHandlers = callHandlers_patched
+        except Exception:
+            pass
+
     def _handle_sigfpe(self, sig, frame):
         self.debugger.set_trace(frame)
 
@@ -555,7 +565,7 @@ class PyzoInterpreter:
             elif guiName == "GTK":
                 self.guiApp = guiintegration.App_gtk()
             else:
-                guiError = "Unkown gui: %s" % guiName
+                guiError = "Unknown gui: %s" % guiName
                 guiName = ""
         except Exception:  # Catch any error
             # Get exception info (we do it using sys.exc_info() because
@@ -1062,8 +1072,8 @@ class PyzoInterpreter:
 
     def ipython_pre_run_cell_hook(self, ipython=None):
         """Hook that IPython calls right before executing code."""
-        self.apply_breakpoints()
         self.debugger.set_on()
+        self.apply_breakpoints()
 
     def ipython_editor_hook(self, ipython, filename, linenum=None, wait=True):
         # Correct line number for cell offset
@@ -1166,58 +1176,29 @@ class PyzoInterpreter:
         try:
             if useLastTraceback:
                 # Get traceback info from buffered
-                type = sys.last_type
+                etype = sys.last_type
                 value = sys.last_value
                 tb = sys.last_traceback
             else:
                 # Get exception information and remove first, since that's us
-                type, value, tb = sys.exc_info()
+                etype, value, tb = sys.exc_info()
                 tb = tb.tb_next
 
                 # Store for debugging, but only store if not in debug mode
                 if not self._dbFrames:
-                    sys.last_type = type
+                    sys.last_type = etype
                     sys.last_value = value
                     sys.last_traceback = tb
 
-            # Get traceback to correct all the line numbers
-            # tblist = list of (filename, line-number, function-name, text)
-            tblist = traceback.extract_tb(tb)
+            lines = traceback.format_exception(etype, value, tb)
 
-            # Get frames
-            frames = []
-            while tb:
-                frames.append(tb.tb_frame)
-                tb = tb.tb_next
-
-            # Walk through the list
-            for i, tbInfo in enumerate(tblist):
-                # Get filename and line number, init example
-                fname, lineno = self.correctfilenameandlineno(tbInfo[0], tbInfo[1])
-                if not isinstance(fname, ustr):
-                    fname = fname.decode("utf-8")
-                example = tbInfo[3]
-                # Reset info
-                tblist[i] = (fname, lineno, tbInfo[2], example)
-
-            # Format list
-            strList = traceback.format_list(tblist)
-            if strList:
-                strList.insert(0, "Traceback (most recent call last):\n")
-            strList.extend(traceback.format_exception_only(type, value))
-
-            # Write traceback
-            for s in strList:
-                self.write(s)
-
-            # Clean up (we cannot combine except and finally in Python <2.5
-            tb = None
-            frames = None
+            for line in lines:
+                line = self.correctfilenameandlineno_on_line(line)
+                self.write(line)
 
         except Exception:
-            type, value, tb = sys.exc_info()
+            etype, value, tb = sys.exc_info()
             tb = None
-            frames = None
             t = "An error occured, but then another one when trying to write the traceback: "
             t += str(value) + "\n"
             self.write(t)
@@ -1237,6 +1218,17 @@ class PyzoInterpreter:
                 pass
         return fname, lineno
 
+    def correctfilenameandlineno_on_line(self, line):
+        match = re.search(file_linenr_pattern, line)
+        if match:
+            i1, i2 = match.start(1), match.end(1)
+            i3, i4 = match.start(2), match.end(2)
+            offset = int(line[i1:i2].lstrip("+"))
+            old_nr = int(line[i3:i4])
+            new_nr = old_nr + offset
+            return line[:i1] + line[i2:i3] + str(new_nr) + line[i4:]
+        return line
+
 
 class ExecutedSourceCollection:
     """Stores the source of executed pieces of code, so that the right
@@ -1272,8 +1264,12 @@ class ExecutedSourceCollection:
                     return linecache._getlines(filename, module_globals)
 
         # Monkey patch
-        linecache._getlines = linecache.getlines
-        linecache.getlines = getlines
+        if not hasattr(linecache, "_getlines"):
+            # Normally, the linecache module is not patched yet.
+            # But when starting the Pyzo kernel in an external shell twice in a row,
+            # we must not patch the module again to prevent infinite recursion.
+            linecache._getlines = linecache.getlines
+            linecache.getlines = getlines
 
         # I hoped this would remove the +lineno for IPython tracebacks,
         # but it doesn't
