@@ -1,13 +1,8 @@
-# Copyright (C) 2016, the Pyzo development team
-#
-# Pyzo is distributed under the terms of the 2-Clause BSD License.
-# The full license can be found in 'license.txt'.
-
-
-import sys, re
+import sys
+import re
 from functools import partial
 
-from pyzo.qt import QtCore, QtGui, QtWidgets  # noqa
+from pyzo.qt import QtCore, QtGui, QtWidgets
 import pyzo
 
 tool_name = pyzo.translate("pyzoInteractiveHelp", "Interactive help")
@@ -74,7 +69,7 @@ Outside such constructs, ``in`` is an operator and its precise meaning depends o
     ),
     "is": pyzo.translate(
         "pyzoInteractiveHelp",
-        """This operator tests for an object’s identity: ``x is y`` is true if and only if ``x`` and ``y`` are the same object.
+        """This operator tests for an object's identity: ``x is y`` is true if and only if ``x`` and ``y`` are the same object.
 
 See also: ``id``""",
     ),
@@ -106,7 +101,7 @@ else:       # optional
 ```
 The expression ``iterable`` is evaluated once; its value should be an iterable object. An iterator is created for the result of that expression. The ``body`` is then executed once for each item provided by the iterator, in the order returned by the iterator: each item in turn is assigned to the ``variable`` and then the ``body`` is executed. When the items are exhausted (which is immediately when the ``iterable`` is empty), the ``suite`` in the ``else`` clause, if present, is executed, and the loop terminates.
 
-A ``break`` statement executed in the first suite terminates the loop without executing the ``else`` clause’s ``suite``. A ``continue`` statement executed in the first suite skips the rest of the ``body`` and continues with the next item, or with the ``else`` clause if there is no next item.
+A ``break`` statement executed in the first suite terminates the loop without executing the ``else`` clause's ``suite``. A ``continue`` statement executed in the first suite skips the rest of the ``body`` and continues with the next item, or with the ``else`` clause if there is no next item.
 
 The for-loop makes assignments to the ``variable``. This overwrites all previous assignments to that variable including those made in the ``body`` of the for-loop:
 ```
@@ -156,7 +151,7 @@ try:
 finally:
     suite
 ```
-If ``critical_section`` raises an exception, the ``suite`` will be executed before aborting the program. The ``suite`` is also executed when no exception is raised. When a ``return``, ``break`` or ``continue`` statement is executed in the ``critical_section``, the ``finally`` clause is also executed ‘on the way out.’ A ``continue`` statement is illegal in the ``finally`` clause.
+If ``critical_section`` raises an exception, the ``suite`` will be executed before aborting the program. The ``suite`` is also executed when no exception is raised. When a ``return``, ``break`` or ``continue`` statement is executed in the ``critical_section``, the ``finally`` clause is also executed 'on the way out.' A ``continue`` statement is illegal in the ``finally`` clause.
 
 Both mechanisms can be combined, the ``finally`` clause coming after the last ``except`` clause or after the optional ``else`` clause.""",
     ),
@@ -189,7 +184,7 @@ else:           # optional
 ```
 This repeatedly tests the ``expression`` and, if it is true, executes the ``body``; if the ``expression`` is false (which may be the first time it is tested) the ``suite`` of the ``else`` clause, if present, is executed and the loop terminates.
 
-A ``break`` statement executed in the ``body`` terminates the loop without executing the ``else`` clause’s ``suite``. A ``continue`` statement executed in the ``body`` skips the rest of the ``body`` and goes back to testing the ``expression``.""",
+A ``break`` statement executed in the ``body`` terminates the loop without executing the ``else`` clause's ``suite``. A ``continue`` statement executed in the ``body`` skips the rest of the ``body`` and goes back to testing the ``expression``.""",
     ),
     "assert": pyzo.translate(
         "pyzoInteractiveHelp",
@@ -263,7 +258,7 @@ See also: ``and``, ``not``, ``True``, ``False``""",
     ),
     "yield": pyzo.translate(
         "pyzoInteractiveHelp",
-        """The ``yield`` expression is used when defining a generator function or an asynchronous generator function and thus can only be used in the body of a function definition. Using a yield expression in a function’s body causes that function to be a generator, and using it in an ``async def`` function’s body causes that coroutine function to be an asynchronous generator.""",
+        """The ``yield`` expression is used when defining a generator function or an asynchronous generator function and thus can only be used in the body of a function definition. Using a yield expression in a function's body causes that function to be a generator, and using it in an ``async def`` function's body causes that coroutine function to be an asynchronous generator.""",
     ),
 }
 
@@ -338,8 +333,9 @@ initText = pyzo.translate(
     "pyzoInteractiveHelp",
     """
 Help information is queried from the current shell
-when moving up/down in the autocompletion list
-and when double clicking on a name.
+automatically when moving up/down in the autocompletion list,
+or manually via the context menu in the shell and editor,
+or via the Workspace tool.
 """,
 )
 
@@ -391,12 +387,20 @@ class PyzoInteractiveHelp(QtWidgets.QWidget):
             PyzoInteractiveHelpHistoryMenu("Forward menu", self, True)
         )
 
+        self._pauseBut = QtWidgets.QToolButton(self)
+        self._pauseBut.setIcon(style.standardIcon(style.StandardPixmap.SP_MediaPause))
+        self._pauseBut.setIconSize(QtCore.QSize(16, 16))
+        self._pauseBut.setCheckable(True)
+        self._pauseBut.setToolTip("Pause automatic help from autocompletion")
+
         # Create options button
         self._options = QtWidgets.QToolButton(self)
         self._options.setIcon(pyzo.icons.wrench)
         self._options.setIconSize(QtCore.QSize(16, 16))
         self._options.setPopupMode(self._options.ToolButtonPopupMode.InstantPopup)
-        self._options.setToolButtonStyle(QtCore.Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
+        self._options.setToolButtonStyle(
+            QtCore.Qt.ToolButtonStyle.ToolButtonTextBesideIcon
+        )
 
         # Create options menu
         self._options._menu = QtWidgets.QMenu()
@@ -405,7 +409,9 @@ class PyzoInteractiveHelp(QtWidgets.QWidget):
         # Create browser
         self._browser = QtWidgets.QTextBrowser(self)
         self._browser_text = initText
-        self._browser.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
+        self._browser.setContextMenuPolicy(
+            QtCore.Qt.ContextMenuPolicy.CustomContextMenu
+        )
         self._browser.customContextMenuRequested.connect(self.showMenu)
 
         # Create two sizers
@@ -415,6 +421,7 @@ class PyzoInteractiveHelp(QtWidgets.QWidget):
         # Put the elements together
         self._sizer2.addWidget(self._backBut, 1)
         self._sizer2.addWidget(self._forwBut, 2)
+        self._sizer2.addWidget(self._pauseBut, 1)
         self._sizer2.addWidget(self._text, 4)
         self._sizer2.addWidget(self._printBut, 0)
         self._sizer2.addWidget(self._options, 3)
@@ -446,6 +453,7 @@ class PyzoInteractiveHelp(QtWidgets.QWidget):
         self._printBut.clicked.connect(self.printDoc)
         self._backBut.clicked.connect(self.goBack)
         self._forwBut.clicked.connect(self.goForward)
+        self._pauseBut.clicked.connect(self.onTogglePause)
         #
         self._options.pressed.connect(self.onOptionsPress)
         self._options._menu.triggered.connect(self.onOptionMenuTiggered)
@@ -546,6 +554,8 @@ class PyzoInteractiveHelp(QtWidgets.QWidget):
         self.queryDoc(addToHist)
 
     def helpFromCompletion(self, name, addToHist=False):
+        if self._pauseBut.isChecked():
+            return
         self.setObjectName(name, addToHist)
 
     def currentHist(self):
@@ -555,13 +565,15 @@ class PyzoInteractiveHelp(QtWidgets.QWidget):
             return None
 
     def addToHist(self, name):
-        if name == self.currentHist():
+        if not name or name == self.currentHist():
             return
         self._history = self._history[: self._histindex + 1]
         self._history.append(name)
         self._histindex = len(self._history) - 1
 
     def restoreCurrent(self):
+        if self._pauseBut.isChecked():
+            return
         self.setObjectName(self.currentHist())
 
     def goBack(self):
@@ -574,6 +586,10 @@ class PyzoInteractiveHelp(QtWidgets.QWidget):
             self._histindex += 1
             self.setObjectName(self.currentHist())
 
+    def onTogglePause(self, pauseActivated):
+        if pauseActivated:
+            self.addToHist(self._text.text())
+
     def printDoc(self):
         """Print the doc for the text in the line edit."""
         # Get name
