@@ -1,9 +1,3 @@
-# -*- coding: utf-8 -*-
-# Copyright (C) 2013, the Pyzo development team
-#
-# Yoton is distributed under the terms of the 2-Clause BSD License.
-# The full license can be found in 'license.txt'.
-
 # This code is loosely based on the event system of Visvis and on the
 # signals system of Qt.
 
@@ -11,7 +5,7 @@
 # here. However, only since Python3.3 is this built-in functionality
 # thread safe. And we need thread safety!
 
-""" Module yoton.events
+"""Module yoton.events
 
 Yoton comes with a simple event system to enable event-driven applications.
 
@@ -40,11 +34,11 @@ class CallableObject(object):
 
     """
 
-    __slots__ = ["_ob", "_func"]  # Use __slots__ to reduce memory footprint
+    __slots__ = ["_func", "_ob"]  # Use __slots__ to reduce memory footprint
 
     def __init__(self, c):
         # Check
-        if not hasattr(c, "__call__"):
+        if not callable(c):
             raise ValueError("Error: given callback is not callable.")
 
         # Store function and object
@@ -116,7 +110,7 @@ class Event(object):
 
     """
 
-    __slots__ = ["_callable", "_args", "_kwargs", "_timeout"]
+    __slots__ = ["_args", "_callable", "_kwargs", "_timeout"]
 
     def __init__(self, callable, *args, **kwargs):
         if isinstance(callable, CallableObject):
